/**
 * \file   my_def.h
 * \brief  common definitions and utilities
 * \author Takahisa Toda
 */
#ifndef MY_DEF_H
#define MY_DEF_H

#include <stdio.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>
#include <time.h>

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

#define inline   inline   __attribute__((always_inline))

#define MY_SUCCESS (0)
#define MY_FAILURE (-1)

#define INFTY  INT32_MAX

typedef intmax_t my_val; //!< label for a decision node. This must be a signed integer!
#define MAX_MYVAL   INTMAX_MAX

/** \brief  Obtain the current time
 *  \return the current time
 *  \see    print_elapsed_time
 */
static inline clock_t my_clock(void)
{
#ifdef TIME_LOG_EXP
  return clock();
#else
  return 0;
#endif /*TIME_LOG_EXP*/
}
/** \brief  Print the elapsed time from a given time.
 *  \param start starting time
 *  \param s message
 *  \see    my_clock
 */
static inline void print_elapsed_time(clock_t start, const char *s)
{
#ifdef TIME_LOG_EXP
  clock_t end = clock();
  printf("\ntimes(%s)\t%.2f sec\n", s, (double)(end - start) / CLOCKS_PER_SEC);
#endif /*TIME_LOG_EXP*/
}

/** \brief  Return a prime number
 *  \param n index
 */
static inline uintmax_t MY_PRIME(int n) 
{ 
  assert(0 <= n && n <= 3);
  return (((n) == 0)? 3343:				\
	  (((n) == 1)? 78498:                           \
	   (((n) == 2)? 664579:				\
	    (((n) == 3)? 5761455:			\
	     137))));
}

#define MAX_RECURSION_DEPTH (8192) //!< Guard againt stack overfow. Adjust this value according to your environment.

#ifdef RECDEPTH_GUARD
#define INC_RECDEPTH(v)							\
  do {									\
    if (v >= MAX_RECURSION_DEPTH)					\
      {									\
	fprintf(stderr, "error:%s:%d:%s: maximum recursion depth exceeded\n", __FILE__, __LINE__, __func__); \
	exit(EXIT_FAILURE);						\
      }									\
    v++;								\
  } while(0)

#define DEC_RECDEPTH(v)							\
  do {									\
    assert(v > 0);							\
    v--;								\
  } while(0)
#else
#define INC_RECDEPTH(v)	do{} while(0)
#define DEC_RECDEPTH(v)	do{} while(0)
#endif /*RECDEPTH_GUARD*/

#define ENSURE_TRUE_WARN(v, msg)                                        \
  do {                                                                  \
    if (!(v))                                                           \
      {                                                                 \
        fprintf(stderr, "warning:%s:%d:%s:%s\n", __FILE__, __LINE__, __func__, msg); \
      }                                                                 \
  } while(0)

#define ENSURE_TRUE(v)							\
  do {									\
    if (!(v))								\
      {									\
	fprintf(stderr, "error:%s:%d:%s failed\n", __FILE__, __LINE__, __func__); \
	exit(EXIT_FAILURE);						\
      }									\
  } while(0)

#define ENSURE_TRUE_MSG(v, msg)						\
  do {									\
    if (!(v))								\
      {									\
	fprintf(stderr, "error:%s:%d:%s:%s\n", __FILE__, __LINE__, __func__, msg); \
	exit(EXIT_FAILURE);						\
      }									\
  } while(0)


#define ENSURE_SUCCESS(v)						\
  do {									\
    if ((v) != MY_SUCCESS)						\
      {									\
	fprintf(stderr, "error:%s:%d:%s failed\n", __FILE__, __LINE__, __func__); \
	exit(EXIT_FAILURE);						\
      }									\
  } while(0)

#define ENSURE_SUCCESS_MSG(v, msg)					\
  do {									\
    if ((v) != MY_SUCCESS)						\
      {									\
	fprintf(stderr, "error:%s:%d:%s:%s\n", __FILE__, __LINE__, __func__, msg); \
	exit(EXIT_FAILURE);						\
      }									\
  } while(0)


#endif /*MY_DEF_H*/
