/** \file   my_tdd.c
 *  \brief  A simple implementation of TDD package
 *  \see T.Toda, Dualization of Boolean functions Using Ternary Decision Diagrams, in Proc. of 13th International Symposium on Artificial Intelligence and Mathematics (ISAIM2014), Florida, USA(2014).
 */
#include <stdio.h>
#include <stdint.h>
#include <assert.h>

#include "my_tdd.h"
#include "my_def.h"
#include "my_opcache.h"

static inline uintmax_t mytdd_calhashval(my_val k, const struct mytdd_node *zero, const struct mytdd_node *neg, const struct mytdd_node *pos, uintmax_t size);
static inline bool mytdd_isvisited(struct mytdd_node *f);
static inline void mytdd_set_visited(struct mytdd_node *f);
static inline void mytdd_clear_visited(struct mytdd_node *f);
static void mytdd_clear_visited_all(struct mytdd_node *f);
static uintmax_t mytdd_size_rec(struct mytdd_node *f);
static void mytdd_create_linkedlist_rec(struct mytdd_node *f, struct mytdd_node **list, uintmax_t *plen, uintmax_t *pmaxlv);
static void mytdd_unique_table_enlarge(void);
static inline struct mytdd_node *mytdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos);
static uintmax_t mytdd_card_rec(struct mytdd_node *f);
static uintmax_t mytdd_totalsize_rec(struct mytdd_node *f);
static struct mytdd_node *mytdd_d2q_rec(my_val maxval, struct mytdd_node *f);

#ifdef MY_DEBUG
#include "my_hash.h"
static int mytdd_check_consistency_rec(struct mytdd_node *f, my_hash *h);
#endif /*MY_DEBUG*/

struct mytdd_node *mytdd_top = NULL; //!< top terminal node
struct mytdd_node *mytdd_bot = NULL; //!< bottom terminal node

static uintmax_t mytdd_excessval_depth_threshold = 0; //!< the number of times that threshold is exceeded.
static bool mytdd_unique_table_exists = false;
static struct mytdd_node **mytdd_unique_table = NULL;
static uintmax_t mytdd_unique_table_size = MYTDD_UNIQUE_TABLE_DEFAULT_SIZE;

static my_opcache *mytdd_opcache = NULL; //!< operation cache
enum MYTDD_OPERATIONS {
  MYTDD_SIZE_OP = 1, //!< id must start from 1.
  MYTDD_CARD_OP,
  MYTDD_TOTALSIZE_OP,
  MYDTDD_INT_OP,
  MYZTDD_DIFF_OP,
  MYTDD_D2Q_OP,
};

static uintmax_t recdepth = 0; //!< recursion depth


/* \brief Create and initialized a hash table for TDD nodes, which is called uniquetable.
 * \param opcache_size the size of a cache for binary operations
 * \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 * \note Nothing is done if a uniquetable already exists.
 */
int mytdd_unique_table_create(uintmax_t opcache_size)
{
#if MYTDD_VARMAX >= INFTY
#error "MYTDD_VARMAX must be less than or equal to INFTY"
#endif

#if MAX_MYVAL < INFTY
#error "INFTY must be less than or equal to MAX_MYVAL"
#endif

  if(mytdd_unique_table_exists) {
    ENSURE_TRUE_WARN(false, "TDD uniquetable already exists");
    return MY_FAILURE;
  }

  mytdd_excessval_depth_threshold = 0;
  mytdd_unique_table_size = MYTDD_UNIQUE_TABLE_DEFAULT_SIZE;

  mytdd_unique_table = (struct mytdd_node**) malloc(sizeof(struct mytdd_node*) * mytdd_unique_table_size);
  ENSURE_TRUE_MSG(mytdd_unique_table != NULL, "memory allocation failed");
  for (uintmax_t i = 0; i < mytdd_unique_table_size; i++) mytdd_unique_table[i] = NULL;

  mytdd_top = (struct mytdd_node*)malloc(sizeof(struct mytdd_node));
  ENSURE_TRUE_MSG(mytdd_top != NULL, "memory allocation failed");
  mytdd_top->k = INFTY;
  mytdd_top->zero = NULL;
  mytdd_top->neg = NULL;
  mytdd_top->pos = NULL;
  mytdd_setaux(0, mytdd_top);
  mytdd_set_visited(mytdd_top);

  mytdd_bot = (struct mytdd_node*)malloc(sizeof(struct mytdd_node));
  ENSURE_TRUE_MSG(mytdd_bot != NULL, "memory allocation failed");
  mytdd_bot->k = INFTY;
  mytdd_bot->zero = NULL;
  mytdd_bot->neg = NULL;
  mytdd_bot->pos = NULL;
  mytdd_setaux(0, mytdd_bot);
  mytdd_set_visited(mytdd_bot);

  mytdd_opcache = opcache_create(opcache_size);
  ENSURE_TRUE_MSG(mytdd_opcache != NULL, "opcache creation failed");

  mytdd_unique_table_exists = true;
  return MY_SUCCESS;
}

/* \brief Free all memory areas used by uniquetable.
 * \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 * \note Nothing is done if a uniquetable does not exist.
 */
int mytdd_unique_table_destroy(void)
{
  if(!mytdd_unique_table_exists) {
    ENSURE_TRUE_WARN(false, "TDD uniquetable no longer exists");
    return MY_FAILURE;
  }

  for(uintmax_t i = 0; i < mytdd_unique_table_size; i++) {
    for(struct mytdd_node *c = mytdd_unique_table[i]; c != NULL; ) {
      struct mytdd_node *t = c->nx;
      free(c);
      c = t;
    }
  }
  free(mytdd_unique_table);
  mytdd_unique_table = NULL;
  free(mytdd_top);
  mytdd_top = NULL;
  free(mytdd_bot);
  mytdd_bot = NULL;

  opcache_destroy(mytdd_opcache);
  mytdd_opcache = NULL;

  mytdd_unique_table_exists = false;
  return MY_SUCCESS;
}

/* \brief Compute the maximum length of a linked list in uniquetable.
 * \return the computed value
 */
uintmax_t mytdd_unique_table_maxdepth(void)
{
  uintmax_t depth = 0;
  for(uintmax_t i = 0; i < mytdd_unique_table_size; i++) {
    uintmax_t j = 0;
    for(struct mytdd_node *c = mytdd_unique_table[i]; c != NULL; c=c->nx) {j++;}
    if(depth < j) {depth = j;}
  }
  return depth;
}

static inline uintmax_t mytdd_calhashval(my_val k, const struct mytdd_node *zero, const struct mytdd_node *neg, const struct mytdd_node *pos, uintmax_t size)
{
  return (k*MY_PRIME(0) + (uintmax_t)zero*MY_PRIME(1) + (uintmax_t)neg*MY_PRIME(2) + (uintmax_t)pos*MY_PRIME(3))%size;
}

/* \brief obtain TDD node with speficied fields.
 * \param k  node label
 * \param zero pointer to ZERO-child (TDD node)
 * \param neg  pointer to NEG-child (TDD node)
 * \param pos  pointer to POS-child (TDD node)
 * \return TDD node
 */
static inline struct mytdd_node *mytdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos)
{
  assert(zero != NULL); assert(neg != NULL); assert(pos != NULL);
  assert(k < zero->k); assert(k < neg->k); assert(k < pos->k);

  uintmax_t depth = 0;
  const uintmax_t val = mytdd_calhashval(k, zero, neg, pos, mytdd_unique_table_size);
  for(struct mytdd_node *n = mytdd_unique_table[val]; n != NULL; n = n->nx) {
    if(n->k == k && n->zero == zero && n->neg == neg && n->pos == pos) {
      return n;
    }
    depth++;
  }
  struct mytdd_node *new = (struct mytdd_node*)malloc(sizeof(struct mytdd_node));
  ENSURE_TRUE_MSG(new != NULL, "memory allocation failed");
  new->k = k; new->zero = zero; new->neg = neg; new->pos = pos;
  mytdd_setaux(0, new);
  mytdd_clear_visited(new);
  new->nx = mytdd_unique_table[val];
  mytdd_unique_table[val] = new;

  if(MYTDD_UNIQUE_TABLE_DEPTH_THRESHOLD <= depth) {
    mytdd_excessval_depth_threshold++;
    if(MYTDD_MAX_EXCESSVAL_DEPTH_THRESHOLD <= mytdd_excessval_depth_threshold) {
      mytdd_unique_table_enlarge();
      mytdd_excessval_depth_threshold = 0;
    }
  }
  return new;
}


/* \brief obtain D-TDD node with speficied fields.
 * \param k  node label
 * \param zero pointer to ZERO-child (D-TDD node)
 * \param neg  pointer to NEG-child (D-TDD node)
 * \param pos  pointer to POS-child (D-TDD node)
 * \return D-TDD node
 */
struct mytdd_node *mydtdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos)
{
  assert(zero != NULL); assert(neg != NULL); assert(pos != NULL);
  assert(k < zero->k); assert(k < neg->k); assert(k < pos->k);

  if(zero == neg && neg == pos) return zero;
  else return mytdd_getnode(k, zero, neg, pos);
}


/* \brief obtain Z-TDD node with speficied fields.
 * \param k  node label
 * \param zero pointer to ZERO-child (Z-TDD node)
 * \param neg  pointer to NEG-child (Z-TDD node)
 * \param pos  pointer to POS-child (Z-TDD node)
 * \return Z-TDD node
 */
struct mytdd_node *myztdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos)
{
  assert(zero != NULL); assert(neg != NULL); assert(pos != NULL);
  assert(k < zero->k); assert(k < neg->k); assert(k < pos->k);

  if(neg == pos && pos == mytdd_bot) return zero;
  else return mytdd_getnode(k, zero, neg, pos);
}

/* \brief obtain Q-TDD node with speficied fields.
 * \param k  node label
 * \param zero pointer to ZERO-child (Q-TDD node)
 * \param neg  pointer to NEG-child (Q-TDD node)
 * \param pos  pointer to POS-child (Q-TDD node)
 * \return Q-TDD node
 * \note Q-TDD means a TDD having node sharing rule, but not having node elimination rule.
 */
struct mytdd_node *myqtdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos)
{
  assert(zero != NULL); assert(neg != NULL); assert(pos != NULL);
  assert(k < zero->k); assert(k < neg->k); assert(k < pos->k);

  return mytdd_getnode(k, zero, neg, pos);
}


/* \brief Clear the visited flags of all nodes below an input node.
 * \param f TDD node
 * \note We assume that the visited flags of all nodes below an input node are set. 
 */
static void mytdd_clear_visited_all(struct mytdd_node *f)
{
  if(f == mytdd_top || f == mytdd_bot) return;
  INC_RECDEPTH(recdepth);
  if(mytdd_isvisited(f)) {
    mytdd_clear_visited_all(f->zero);
    mytdd_clear_visited_all(f->neg);
    mytdd_clear_visited_all(f->pos);
    mytdd_clear_visited(f);
  }
  DEC_RECDEPTH(recdepth);  
}

/* \brief Answer whether the visited flag of a node is set.
 * \param f TDD node
 * \return true if visited; otherwise, false.
 */
static inline bool mytdd_isvisited(struct mytdd_node *f)
{
  return f->visited;
}


/* \brief Set a visited flag.
 * \param f TDD node
 * \note Do not forget to clear visited-flags!
 */
static inline void mytdd_set_visited(struct mytdd_node *f)
{
  f->visited = true;
}

/* \brief Clear a visited flag.
 * \param f TDD node
 */
static inline void mytdd_clear_visited(struct mytdd_node *f)
{
  f->visited = false;
}


#ifdef MY_DEBUG
/* \brief Check consisitency of all nodes below an input node.
 * \param f TDD node
 * \return MY_SUCCESS if no problems; otherwise, MY_FAILURE.
 */
int mytdd_check_consistency(struct mytdd_node *f)
{
  assert(f != NULL);
  my_hash *h = ht_create(HASH_DEFAULT_SIZE);
  ENSURE_TRUE_MSG(h != NULL, "hash creation failed");
  int res = mytdd_check_consistency_rec(f, h);
  ht_destroy(h);
  return res;
}


static int mytdd_check_consistency_rec(struct mytdd_node *f, my_hash *h)
{
  if(f == mytdd_top || f == mytdd_bot) {
    if(f->k == INFTY && f->zero == NULL && f->neg == NULL && f->pos == NULL && f->nx == NULL && mytdd_isvisited(f)) {
      return MY_SUCCESS;
    } else {
      ENSURE_TRUE_WARN(f->k    == INFTY, "the k field of a constant node is inconsistent");
      ENSURE_TRUE_WARN(f->zero == NULL, "the zero field of a constant node is inconsistent");
      ENSURE_TRUE_WARN(f->neg  == NULL, "the neg field of a constant node is inconsistent");
      ENSURE_TRUE_WARN(f->pos  == NULL, "the pos field of a constant node is inconsistent");
      ENSURE_TRUE_WARN(f->nx   == NULL, "the nx field of a constant node is inconsistent");
      ENSURE_TRUE_WARN(mytdd_isvisited(f), "the visited field of a constant node is inconsistent");
      return MY_FAILURE;
    }
  }

  int res; int result;
  if((res = ht_search((uintptr_t)f, (uintptr_t*)&result, h)) == MY_SUCCESS) return result;

  INC_RECDEPTH(recdepth);
  result = mytdd_check_consistency_rec(f->zero, h);
  if(result != MY_SUCCESS) {DEC_RECDEPTH(recdepth); return result;}
  result = mytdd_check_consistency_rec(f->neg, h);
  if(result != MY_SUCCESS) {DEC_RECDEPTH(recdepth); return result;}
  result = mytdd_check_consistency_rec(f->pos, h);
  if(result != MY_SUCCESS) {DEC_RECDEPTH(recdepth); return result;}
  
  if(0 < f->k && f->k <= MYTDD_VARMAX && f->zero != NULL && f->neg != NULL && f->pos != NULL && !mytdd_isvisited(f)) {
    result = MY_SUCCESS;
  } else {
    ENSURE_TRUE_WARN(0 < f->k && f->k <= MYTDD_VARMAX, "the k field of an internal node is inconsistent");
    ENSURE_TRUE_WARN(f->zero != NULL, "the zero field of an internal node is inconsistent");
    ENSURE_TRUE_WARN(f->neg  != NULL, "the neg field of an internal node is inconsistent");
    ENSURE_TRUE_WARN(f->pos  != NULL, "the pos field of an internal node is inconsistent");
    ENSURE_TRUE_WARN(!mytdd_isvisited(f), "the visited field of an internal node is inconsistent");
    result = MY_FAILURE;
  }
  
  ht_add((uintptr_t)f, (uintptr_t)result, h);

  DEC_RECDEPTH(recdepth);
  return result;
}
#endif /*MY_DEBUG*/

/* \brief Convert D-TDD to Q-TDD
 * \param f D-TDD
 * \return Q-TDD
 * \note Q-TDD means a TDD having node sharing rule, but not having node elimination rule.
 */
struct mytdd_node *mytdd_d2q(my_val maxval, struct mytdd_node *f)
{
  if(f == mytdd_top || f == mytdd_bot) {
    for(my_val i = maxval; i > 0; i--) {
      f = myqtdd_getnode(i, f, f, f);
    }
  } else {
    struct mytdd_node *g = mytdd_d2q_rec(maxval, f);
    for(my_val i = f->k - 1; i > 0; i--) {
      g = myqtdd_getnode(i, g, g, g);
    }
    f = g;
  }
  return f;
}


static struct mytdd_node *mytdd_d2q_rec(my_val maxval, struct mytdd_node *f)
{
  if(f == mytdd_top || f == mytdd_bot) return f;

  int res; struct mytdd_node *r;
  if((res = opcache_search(MYTDD_D2Q_OP, (uintptr_t)maxval, (uintptr_t)f, (uintptr_t*)&r, mytdd_opcache)) == MY_SUCCESS) return r;
  
  INC_RECDEPTH(recdepth);
  struct mytdd_node *r0 = mytdd_d2q_rec(maxval, f->zero);
  my_val lv;
  if(r0 == mytdd_top || r0 == mytdd_bot) lv = maxval;
  else lv = r0->k - 1;
  for(my_val i = lv; i > f->k; i--) {
    r0 = myqtdd_getnode(i, r0, r0, r0);
    ENSURE_TRUE_MSG(r0 != NULL, "TDD operation failed");
  }

  struct mytdd_node *r1 = mytdd_d2q_rec(maxval, f->neg);
  if(r1 == mytdd_top || r1 == mytdd_bot) lv = maxval;
  else lv = r1->k - 1;
  for(my_val i = lv; i > f->k; i--) {
    r1 = myqtdd_getnode(i, r1, r1, r1);
    ENSURE_TRUE_MSG(r1 != NULL, "TDD operation failed");
  }

  struct mytdd_node *r2 = mytdd_d2q_rec(maxval, f->pos);
  if(r2 == mytdd_top || r2 == mytdd_bot) lv = maxval;
  else lv = r2->k - 1;
  for(my_val i = lv; i > f->k; i--) {
    r2 = myqtdd_getnode(i, r2, r2, r2);
    ENSURE_TRUE_MSG(r2 != NULL, "TDD operation failed");
  }

  r = myqtdd_getnode(f->k, r0, r1, r2);
  ENSURE_TRUE_MSG(r != NULL, "TDD operation failed");
  opcache_add(MYTDD_D2Q_OP, (uintptr_t)maxval, (uintptr_t)f, (uintptr_t)r, mytdd_opcache);
  DEC_RECDEPTH(recdepth);
  return r;
}


/* \brief compute the number of internal TDD nodes below an input TDD node.
 * \param f a TDD node
 * \return the computed size
 * \note
 * - this function does not count terminal nodes.
 */
uintmax_t mytdd_size(struct mytdd_node *f)
{
#ifdef MY_DEBUG
  ENSURE_SUCCESS_MSG(mytdd_check_consistency(f), "inconsistency was detected");
#endif /*MY_DEBUG*/
  int res; uintmax_t size;
  if((res = opcache_search(MYTDD_SIZE_OP, (uintptr_t)f, 0, (uintptr_t*)&size, mytdd_opcache)) == MY_SUCCESS) return size;
  size = mytdd_size_rec(f);
  opcache_add(MYTDD_SIZE_OP, (uintptr_t)f, 0, (uintptr_t)size, mytdd_opcache);
  mytdd_clear_visited_all(f);
  return size;
}

static uintmax_t mytdd_size_rec(struct mytdd_node *f)
{
  assert(f != NULL);
  if(mytdd_isvisited(f)) return 0;

  INC_RECDEPTH(recdepth);
  uintmax_t size  = mytdd_size_rec(f->zero) + mytdd_size_rec(f->neg) + mytdd_size_rec(f->pos) + 1;
  mytdd_set_visited(f);
  DEC_RECDEPTH(recdepth);
  return size;
}

/* \brief Compute the number of paths from f to top terminal nodes. After execution, for each node below f, the number of paths from the node to top terminal nodes are set to the aux field.
 * \param f Z-TDD or a non-constant Q-TDD
 * \return the computed number
 * \note 
 * - input TDD must not be a D-TDD.
 * - Q-TDD must not be constant.
 * - Q-TDD means a TDD having node sharing rule, but not having node elimination rule.
 */
uintmax_t mytdd_card(struct mytdd_node *f)
{
#ifdef MY_DEBUG
  ENSURE_SUCCESS_MSG(mytdd_check_consistency(f), "inconsistency was detected");
#endif /*MY_DEBUG*/
  uintmax_t card = mytdd_card_rec(f);
  mytdd_clear_visited_all(f);
  return card;
}

static uintmax_t mytdd_card_rec(struct mytdd_node *f)
{
  if(f == mytdd_top) {mytdd_setaux(1,f); return 1;}
  if(f == mytdd_bot) {mytdd_setaux(0,f); return 0;}

  if(mytdd_isvisited(f)) return (uintmax_t)(f->aux);
 
  INC_RECDEPTH(recdepth);
  uintmax_t card0 = mytdd_card_rec(f->zero);
  uintmax_t card1 = mytdd_card_rec(f->neg);
  uintmax_t card2 = mytdd_card_rec(f->pos);
  ENSURE_TRUE_WARN(card0 < UINTMAX_MAX - card1, "integer overflow");
  ENSURE_TRUE_WARN(card0 + card1 < UINTMAX_MAX - card2, "integer overflow");
  uintmax_t card  = card0 + card1 + card2;
  mytdd_setaux((uintptr_t)card, f);
  mytdd_set_visited(f);
  DEC_RECDEPTH(recdepth);
  return card;
}

/* \brief compute the total size of f when f is explicitely represented.
 * \param f Z-TDD or non-constant Q-TDD
 * \return the computed size
 * - input TDD must not be a D-TDD.
 * - input TDD must be a non-constant node.
 * - Q-TDD means a TDD having node sharing rule, but not having node elimination rule.
 */
uintmax_t mytdd_totalsize(struct mytdd_node *f)
{
  mytdd_card(f); // compute the cardinality of each node and set it to the aux field in advance.
  return mytdd_totalsize_rec(f);
}

static uintmax_t mytdd_totalsize_rec(struct mytdd_node *f)
{
  if(f == mytdd_top || f == mytdd_bot) return 0;

  int res; uintmax_t size;
  if((res = opcache_search(MYTDD_TOTALSIZE_OP, (uintptr_t)f, 0, (uintptr_t*)&size, mytdd_opcache)) == MY_SUCCESS) return size;

  INC_RECDEPTH(recdepth);
  uintmax_t size0 = mytdd_totalsize_rec(f->zero);
  uintmax_t size1 = mytdd_totalsize_rec(f->neg);
  uintmax_t size2 = mytdd_totalsize_rec(f->pos);
  ENSURE_TRUE_WARN(size0 < UINTMAX_MAX - size1, "integer overflow");
  ENSURE_TRUE_WARN(size0 + size1 < UINTMAX_MAX - size2, "integer overflow");
  size = size0 + size1 + size2;
  uintmax_t card1 = f->neg->aux;
  uintmax_t card2 = f->pos->aux;
  ENSURE_TRUE_WARN(size < UINTMAX_MAX - card1, "integer overflow");
  ENSURE_TRUE_WARN(size + card1 < UINTMAX_MAX - card2, "integer overflow");
  size += card1 + card2;
  opcache_add(MYTDD_TOTALSIZE_OP, (uintptr_t)f, 0, (uintptr_t)size, mytdd_opcache);
  DEC_RECDEPTH(recdepth);
  return size;
}

/* \brief Compute the difference of two Z-TDDs
 * \param f Z-TDD
 * \param g Z-TDD
 * \return the computed Z-TDD
 */
struct mytdd_node *myztdd_diff(struct mytdd_node *f, struct mytdd_node *g)
{
  if(f == g || f == mytdd_bot) return mytdd_bot;
  if(g == mytdd_bot) return f;
  if(g == mytdd_top) {
    INC_RECDEPTH(recdepth);
    struct mytdd_node *r = myztdd_getnode(f->k, myztdd_diff(f->zero, g), f->neg, f->pos);
    ENSURE_TRUE_MSG(r != NULL, "Z-TDD operation failed");
    DEC_RECDEPTH(recdepth);
    return r;
  }
  if(f == mytdd_top) {
    INC_RECDEPTH(recdepth);
    struct mytdd_node *r = myztdd_diff(f, g->zero);
    DEC_RECDEPTH(recdepth);
    return r;
  }

  int res; struct mytdd_node *r;
  if((res = opcache_search(MYZTDD_DIFF_OP, (uintptr_t)f, (uintptr_t)g, (uintptr_t*)&r, mytdd_opcache)) == MY_SUCCESS) return r;

  INC_RECDEPTH(recdepth);
  if      (f->k > g->k) r = myztdd_diff(f, g->zero);
  else if (f->k < g->k) r = myztdd_getnode(f->k, myztdd_diff(f->zero, g), f->neg, f->pos);
  else                  r = myztdd_getnode(f->k, myztdd_diff(f->zero, g->zero), myztdd_diff(f->neg, g->neg), myztdd_diff(f->pos, g->pos));
  ENSURE_TRUE_MSG(r != NULL, "Z-TDD operation failed");
  
  opcache_add(MYZTDD_DIFF_OP, (uintptr_t)f, (uintptr_t)g, (uintptr_t)r, mytdd_opcache);
  DEC_RECDEPTH(recdepth);
  return r;
}


/* \brief Create an array and set all nodes in f in the array in decreasing order with respect to their node lables.
 * \param f TDD
 * \see mytdd_destroy_array_with_nodes_maxlenwise
 */
struct mytdd_nodearray *mytdd_create_array_with_nodes_levelwise(my_val maxval, struct mytdd_node *f)
{
#ifdef MY_DEBUG
  mytdd_check_consistency(f);
#endif /*MY_DEBUG*/

  if(f == mytdd_top || f == mytdd_bot) {
    ENSURE_TRUE_WARN(false, "constant TDD was given");
    return NULL;
  }

  assert(maxval > 0);
  struct mytdd_node **list = (struct mytdd_node **)malloc(sizeof(struct mytdd_node*) * (maxval+1));
  ENSURE_TRUE_MSG(list != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i <= maxval; i++) list[i] = NULL;

  uintmax_t len   = 0;
  uintmax_t maxlv = 0;
  mytdd_create_linkedlist_rec(f, list, &len, &maxlv);

  struct mytdd_node **a = (struct mytdd_node**)malloc(sizeof(struct mytdd_node*) * len);
  ENSURE_TRUE_MSG(a != NULL, "memory allocation failed");

  uintmax_t nlevels = 0;
  for(my_val i = 1; i <= maxval; i++) {
    if(list[i] != NULL) nlevels++;
  }

  uintmax_t *b = (uintmax_t*)malloc(sizeof(uintmax_t) * (nlevels+1));
  ENSURE_TRUE_MSG(b != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i <= nlevels ; i++) b[i] = 0;

  uintmax_t a_pos = 0;
  uintmax_t b_pos = 0;
  for(uintmax_t i = maxlv; i > 0; i--) {
    if(list[i] != NULL) {b_pos++;}
    for(struct mytdd_node *f = list[i]; f != NULL;) {
      b[b_pos]++;
      struct mytdd_node *nx = (struct mytdd_node*)f->aux;
      assert(a_pos < len);
      a[a_pos++] = f;
      mytdd_setaux((uintptr_t)NULL, f);
      mytdd_clear_visited(f);
      f = nx;
    }
  }

  free(list);

  struct mytdd_nodearray *na = (struct mytdd_nodearray*)malloc(sizeof(struct mytdd_nodearray));
  ENSURE_TRUE_MSG(na != NULL, "memory allocation failed");
  na->a = a;
  na->b = b;
  na->len = len;
  na->nlevels = nlevels;
 
  return na;
}


void mytdd_destroy_array_with_nodes_levelwise(struct mytdd_nodearray *na)
{
  free(na->a);
  free(na->b);
  free(na);
}

static void mytdd_create_linkedlist_rec(struct mytdd_node *f, struct mytdd_node **list, uintmax_t *plen, uintmax_t *pmaxlv)
{
  if(mytdd_isvisited(f)) return;

  INC_RECDEPTH(recdepth);
  mytdd_create_linkedlist_rec(f->zero, list, plen, pmaxlv);
  mytdd_create_linkedlist_rec(f->neg,  list, plen, pmaxlv);
  mytdd_create_linkedlist_rec(f->pos,  list, plen, pmaxlv);
  mytdd_setaux((uintptr_t)list[f->k], f);
  list[f->k] = f;
  (*plen)++;
  if((*pmaxlv) < f->k) (*pmaxlv) = f->k;
  mytdd_set_visited(f);
  DEC_RECDEPTH(recdepth);
}

/* \brief Compute the intersection of two D-TDDs
 * \param f D-TDD
 * \param g D-TDD
 * \return the computed D-TDD
 */
struct mytdd_node *mydtdd_intersect(struct mytdd_node *f, struct mytdd_node *g)
{
  if(f == mytdd_bot || g == mytdd_bot) return mytdd_bot;
  if(f == mytdd_top) return g;
  if(g == mytdd_top) return f;

  int res; struct mytdd_node *r;
  if((res = opcache_search(MYDTDD_INT_OP, (uintptr_t)f, (uintptr_t)g, (uintptr_t*)&r, mytdd_opcache)) == MY_SUCCESS) return r;

  struct mytdd_node *h0; struct mytdd_node *hm; struct mytdd_node *hp;
  INC_RECDEPTH(recdepth);
  if      (f->k > g->k) {
    h0 = mydtdd_intersect(f, g->zero);
    hm = mydtdd_intersect(f, g->neg);
    hp = mydtdd_intersect(f, g->pos);
  }
  else if (f->k < g->k) {DEC_RECDEPTH(recdepth); return mydtdd_intersect(g,f);}
  else {
    h0 = mydtdd_intersect(f->zero, g->zero);
    hm = mydtdd_intersect(f->neg, g->neg);
    hp = mydtdd_intersect(f->pos, g->pos);
  }
  r = mydtdd_getnode(g->k, h0, hm, hp);
  ENSURE_TRUE_MSG(r != NULL, "TDD operation failed");
  
  opcache_add(MYDTDD_INT_OP, (uintptr_t)f, (uintptr_t)g, (uintptr_t)r, mytdd_opcache);
  DEC_RECDEPTH(recdepth);
  return r;
}



/* \brief Enlarge the unique table of TDDs.
 */
void mytdd_unique_table_enlarge(void)
{
  const uintmax_t oldsiz = mytdd_unique_table_size;
  ENSURE_TRUE_MSG(oldsiz < UINTMAX_MAX / 4, "uniquetable size overflow");
  const uintmax_t newsiz = oldsiz << 2;
#ifdef LOG_INFO
  printf("TDD uniquetable resizing:%ju -> %ju\n", oldsiz, newsiz);
#endif /*LOG_INFO*/
  struct mytdd_node **oldtable = mytdd_unique_table;
  struct mytdd_node **newtable = (struct mytdd_node**) malloc(sizeof(struct mytdd_node*) * newsiz);
  ENSURE_TRUE_MSG(newtable != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i < newsiz; i++) newtable[i] = NULL;
  for(uintmax_t i = 0; i < oldsiz; i++) {
    for(struct mytdd_node *c = oldtable[i]; c != NULL;) {
      struct mytdd_node *nextnode = c->nx;
      const uintmax_t val = mytdd_calhashval(c->k, c->zero, c->neg, c->pos, newsiz);
      c->nx = newtable[val];
      newtable[val] = c;
      c = nextnode;
    }
  }
  free(oldtable); oldtable = NULL;
  mytdd_unique_table_size = newsiz;
  mytdd_unique_table = newtable;

  opcache_enlarge(mytdd_opcache);
}



int tdd_export_dot(my_val maxval, struct mytdd_node *f, FILE *out)
{  
  if(f == NULL || f == mytdd_top || f == mytdd_bot) {
    ENSURE_TRUE_WARN(false, "invalid input");
    exit(EXIT_FAILURE);
  }

  int res = fseek(out, 0L, SEEK_SET);
  ENSURE_SUCCESS(res);

  struct mytdd_nodearray *na = mytdd_create_array_with_nodes_levelwise(maxval, f);
  uintmax_t index = 2; // bottom node has index 0, and top node has index 1.
  mytdd_setaux(0, mytdd_bot);
  mytdd_setaux(1, mytdd_top);
  fprintf(out, "digraph tdd {\n");
  fprintf(out, "{rank = same; 0 1}\n");
  for(uintmax_t i = 1; i <= na->nlevels; i++) {
    fprintf(out, "{rank = same;");
    for(uintmax_t j = 1; j <= na->b[i]; j++) fprintf(out, " %ju", index++);
    fprintf(out, "}\n");
  }
  assert(index == na->len + 2);
  for(uintmax_t i = 0; i < na->len; i++) {
    fprintf(out, "%ju [label = %jd];\n",           i+2, na->a[i]->k);
    fprintf(out, "%ju -> %ju ;\n",                 i+2, na->a[i]->pos->aux);
    fprintf(out, "%ju -> %ju [style = dashed];\n", i+2, na->a[i]->neg->aux);
    fprintf(out, "%ju -> %ju [style = dotted];\n", i+2, na->a[i]->zero->aux);
    mytdd_setaux(i+2, na->a[i]);
  }
  fprintf(out, "1 [shape=box];\n");
  fprintf(out, "0 [shape=box];\n");
  fprintf(out, "}\n");

  mytdd_destroy_array_with_nodes_levelwise(na);
  return MY_SUCCESS;
}



