/**
 * \file  ssort2.c
 * \brief Implementation of a sort algorithm for a hypergraph with signed entries.
 * \note  We modified the multikey quicksort by Bentley and Sedgewick: the original source code can be obtained from http://www.cs.princeton.edu/~rs/strings .
 */


#include <stdlib.h>
#include <assert.h>

#include "ssort2.h"
#include "my_random.h"

#ifdef i2c
#error "The macro i2c is already defined."
#else
#define i2c(i) x[i][depth]
#endif /*i2c*/

static inline uintmax_t min(uintmax_t a, uintmax_t b)
{
  return ((a)<=(b) ? (a) : (b));
}

static inline void swap(uintmax_t a, uintmax_t b, my_val *x[])
{
  my_val *t=x[a];
  x[a]=x[b]; x[b]=t;
}

static inline void vecswap(uintmax_t i, uintmax_t j, uintmax_t n, my_val *x[])
{   while (n-- > 0) {
        swap(i, j, x);
        i++;
        j++;
    }
}

static uintmax_t recdepth = 0; //!< recursion depth

static void ssort2(my_val *x[], uintmax_t n, uintmax_t depth)
{   
  uintmax_t a, b, c, d, m;
  my_val r, rr, v;
  if (n <= 1) return;
  INC_RECDEPTH(recdepth);
  a = my_rand(n);
  swap(0, a, x);
  v = i2c(0);
  a = b = 1;
  c = d = n-1;
  for (;;) {
    while (b <= c && (r = abs(i2c(b))-abs(v)) <= 0) {
      if ((rr = i2c(b)-v) == 0) { swap(a, b, x); a++;}
      if (r == 0 && rr > 0 ) break;
      b++;
    }
    while (b <= c && (r = abs(i2c(c))-abs(v)) >= 0) {
      if ((rr = i2c(c)-v) == 0) { swap(c, d, x); d--;}
      if (r == 0 && rr < 0) break;
      c--;
    }
    if (b > c) break;
    swap(b, c, x);
    b++;
    c--;
  }
  m = min(a, b-a);     vecswap(0, b-m, m, x);
  m = min(d-c, n-d-1); vecswap(b, n-m, m, x);
  m = b-a; ssort2(x, m, depth);
  assert(i2c(m) == v);
  if (i2c(m) != INFTY) {
    ssort2(x + m, a + n-d-1, depth+1);
    m = d-c; ssort2(x + n - m, m, depth);
  } else {
    assert(d-c == 0);
  }
  DEC_RECDEPTH(recdepth);
}

void ssort2main(my_val *x[], uintmax_t n)
{assert(x != NULL); ssort2(x, n, 0);}

