/** \file   tdd_min.c
 *  \author Takahisa Toda
 *  \see T.Toda, Dualization of Boolean functions Using Ternary Decision Diagrams, in Proc. of 13th International Symposium on Artificial Intelligence and Mathematics (ISAIM2014), Florida, USA(2014).
 */


#include "tdd_min.h"
#include "my_hash.h"

static my_ztddp min_d2z_rec(my_dtddp f, my_hash *h);


#ifdef SIZE_LOG_EXP
static uintmax_t maxztddsize = 0; //!< maximum size of Z-TDDs
#endif /*SIZE_LOG_EXP*/


/** \brief Construct the Z-TDD that represents all minimal sets in a D-TDD.
 *  \param f    D-TDD
 *  \return the Z-TDD contructed.
 */
my_ztddp min_d2z(my_dtddp f)
{
  my_hash *h = ht_create(HASH_DEFAULT_SIZE); /*adjust HASH_DEFAULT_SIZE if necessary.*/
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

#ifdef SIZE_LOG_EXP
  maxztddsize = 0;
#endif /*SIZE_LOG_EXP*/

  my_ztddp r = min_d2z_rec(f, h);
  ENSURE_TRUE(r != ZTDD_NULL);

#ifdef SIZE_LOG_EXP
  printf("maxztddsize:\t%ju\n", maxztddsize);
#endif /*SIZE_LOG_EXP*/
#ifdef MY_DEBUG
  printf("maxhashdepth\t%ju\n", ht_maxdepth(h));
#endif /*MY_DEBUG*/

  ht_destroy(h);
  return r;
}

static my_ztddp min_d2z_rec(my_dtddp f, my_hash *h)
{
  if(f == dtdd_top()) return ztdd_top();
  if(f == dtdd_bot()) return ztdd_bot();

  int res; my_ztddp r;
  if((res = ht_search((uintptr_t)f, (uintptr_t*)&r, h)) == MY_SUCCESS) return r;

  my_ztddp rzero = min_d2z_rec(dtdd_ZERO(f), h);  
  my_ztddp rneg  = ztdd_diff(min_d2z_rec(dtdd_NEG(f), h), rzero);
  ENSURE_TRUE_MSG(rneg != ZTDD_NULL, "Z-TDD operation failed");
  my_ztddp rpos  = ztdd_diff(min_d2z_rec(dtdd_POS(f), h), rzero);
  ENSURE_TRUE_MSG(rpos != ZTDD_NULL, "Z-TDD operation failed");

  r = ztdd_getnode(dtdd_label(f), rzero, rneg, rpos);
  ENSURE_TRUE_MSG(r != ZTDD_NULL, "Z-TDD operation failed");

  ht_add((uintptr_t)f, (uintptr_t)r, h);
  
#ifdef SIZE_LOG_EXP
  const uintmax_t outsize = ztdd_size(r);
  if(maxztddsize < outsize) maxztddsize = outsize;
#endif /*SIZE_LOG_EXP*/

  return r;
}

