/** \file   tdd_trans.c
 *  \author Takahisa Toda
 *  \see T.Toda, Dualization of Boolean functions Using Ternary Decision Diagrams, in Proc. of 13th International Symposium on Artificial Intelligence and Mathematics (ISAIM2014), Florida, USA(2014).
 */

#include "tdd_trans.h"
#include "my_hash.h"

static my_dtddp trans_z2d_rec(my_ztddp f, my_hash *h);


#ifdef SIZE_LOG_EXP
static uintmax_t maxdtddsize = 0; //!< maximum size of D-TDDs
#endif /*SIZE_LOG_EXP*/


/** \brief Construct the D-TDD that represents all transversals for a Z-TDD.
 *  \param f    Z-TDD
 *  \return the D-TDD contructed.
 */
my_dtddp trans_z2d(my_ztddp f)
{
  my_hash *h = ht_create(HASH_DEFAULT_SIZE); /*adjust HASH_DEFAULT_SIZE if necessary.*/
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

#ifdef SIZE_LOG_EXP
  maxdtddsize = 0;
#endif /*SIZE_LOG_EXP*/

  my_dtddp r = trans_z2d_rec(f, h);
  ENSURE_TRUE(r != DTDD_NULL);

#ifdef SIZE_LOG_EXP
  printf("maxdtddsize:\t%ju\n", maxdtddsize);
#endif /*SIZE_LOG_EXP*/
#ifdef MY_DEBUG
  printf("maxhashdepth\t%ju\n", ht_maxdepth(h));
#endif /*MY_DEBUG*/

  ht_destroy(h);
  return r;
}

static my_dtddp trans_z2d_rec(my_ztddp f, my_hash *h)
{
  if(f == ztdd_top()) return dtdd_bot();
  if(f == ztdd_bot()) return dtdd_top();

  int res; my_dtddp r;
  if((res = ht_search((uintptr_t)f, (uintptr_t*)&r, h)) == MY_SUCCESS) return r;

  my_dtddp gzero = trans_z2d_rec(ztdd_ZERO(f), h);  
  my_dtddp gneg  = trans_z2d_rec(ztdd_NEG(f), h);
  my_dtddp gpos  = trans_z2d_rec(ztdd_POS(f), h);

  my_dtddp tpos = dtdd_intersect(gzero, gneg); 
  ENSURE_TRUE_MSG(tpos != DTDD_NULL, "D-TDD operation failed");
  my_dtddp tneg = dtdd_intersect(gzero, gpos); 
  ENSURE_TRUE_MSG(tneg != DTDD_NULL, "D-TDD operation failed");
  my_dtddp tzero = dtdd_intersect(tpos, gpos); 
  ENSURE_TRUE_MSG(tzero != DTDD_NULL, "D-TDD operation failed");

  r = dtdd_getnode(ztdd_label(f), tzero, tneg, tpos);
  ENSURE_TRUE_MSG(r != DTDD_NULL, "D-TDD operation failed");

  ht_add((uintptr_t)f, (uintptr_t)r, h);
  
#ifdef SIZE_LOG_EXP
  const uintmax_t outsize = dtdd_size(r);
  if(maxdtddsize < outsize) maxdtddsize = outsize;
#endif /*SIZE_LOG_EXP*/

  return r;
}

